'-----------------------------------------------------------------------
' Program :                     TUKIN.PRG
' Author  :                     Mirko B.
' Description :                 User Kinematics example
'                               User Kinematics on example of SCORBOT kinematics.
'                               Shows how to configure a user-kinematics group and how to 
'                               test it. should be started after User.o is loaded and the 
'                               USR_SCORBOT was called.
' Input parameter(s)  :
' Output parameter(s) :
' Global & System variable(s) : UKIN, group, model = 5, of XYZPR
' History of modification(s) :
'-----------------------------------------------------------------------
dim shared prevvel as joint of xyzrp            ' variable used as previous velocity for VelCmd computation
dim shared prevpos as joint of xyzrp             ' variable used as previous position for VelCmd computation
dim shared period as long = 4                    ' value of SERCOS period in ms
dim shared eVcmd as error "VCMD invalid!"        ' Local user error
dim shared eKin  as error "Kinematics mismatch!" ' Local user error
dim shared GlobalPositionErrorDelay as double = 4 ' depends if it is simulated or ECAT or CAN system
program
   dim cpos as location of xyzrp
   dim jpos as joint    of xyzrp

   ' Event for checking VelCmd computation, if triggered something is wrong with vcmd
   onevent vchk CheckVel 
       throw eVcmd 
   end onevent

   Print "User Kinemaitcs version = ";user USR_VER(0,0,0)
   ?user USR_SCORBOT(elementid,0,0)

   ' Group setup (SIMULATION), nothing special any group setup procedure can be used....
   ser.cycletime = period*1000 ' SERCOS period
   call SetAxis(a1,165,1600)  
   call SetAxis(a2,127,1200)  
   call SetAxis(a3,150,1500)  
   call SetAxis(a4,225,2200)  
   call SetAxis(a5,330,3300)  
   call Setgroup(Ukin)
   ' ....................................................................................

   
   with Ukin
    
     ' Event initialization and start
     prevpos = pcmd ' initial values
     prevvel = vcmd
     eventon vchk

     ' Test recording
     record tukin.rec 10000 gap = 1 recdata = setpoint{1},setpoint{2},setpoint{3},a1.pcmd,a2.pcmd, a3.pcmd
      recordon
  
     attach

        ' Test Inverse and direct:
        jpos = {0, -45, 45, -90, 0}
        cpos = ToCart(ukin,jpos)
        if ( Dist(jpos,ToJoint(ukin,cpos,0)) > 0.0001 ) then
          throw eKin
        end if

        ' Enable and start some different motion types, for different kinematics different points could be needed!
        en = 1
  
        move  jpos                   ' Initial position (JOINT INTERPOLATION)
        move  cpos                   ' Just checking inverse kinematics, same point as above
        delay 120
  
        moves #{600,0,380,180,90}    ' Correct it   (CARTESIAN INTERPOLATION ...)
        moves #{620,0,380,180,90}    ' Forward
        moves #{450,0,380,180,90}    ' Back
        moves #{500,0,280,180,90}    ' Down
        moves #{500,0,480,180,90}    ' Up  
        moves #{500,100,480,180,90}  ' Left
        moves #{500,-100,480,180,90} ' Right

     detach
     recordclose
     eventoff vchk
   end with
   Print "OK, user kinematics seems to be properly installed"
end program

sub SetAxis( ax as generic axis, byval v as double, byval a as double)
  with ax
  attach
    en = 0
    simulated = 0
    smooth = -1
    pfac = 65536
    vfac = pfac/1000
    afac = vfac/1000
    jfac = afac/1000
    vmax = v
    amax = a
    jmax    = 20*amax
    vospd   = 1.2*vmax
    vcruise = 0.5*vmax
    acc = amax
    dec = amax
    pemax = 1
    PositionRolloverEnable = off
    simulated = 1
    PositionErrorDelay = GlobalPositionErrorDelay

    detach
  end with  
end sub

sub SetGroup (gr as generic group)
  with gr
    attach
    en=0
   smooth = -1
  vfac = 1/1000
  afac = vfac/1000
  jfac = afac/1000
  vmax = max(max(max(a1.vmax,a2.vmax),max(a3.vmax,a4.vmax)),a5.vmax)
  amax = max(max(max(a1.amax,a2.amax),max(a3.amax,a4.amax)),a5.amax)
  dmax = amax
  jmax = max(max(max(a1.jmax,a2.jmax),max(a3.jmax,a4.jmax)),a5.jmax)

    coupled = 0
  abs = 1
    acmd = 0
    ecmd = 0
    wcmd = 0
  configgroup 

  vcruise = 0.5 * Vmax
  acc = amax
  dec = dmax
  jmax = 10*amax
  jerk= jmax


  velocityMaxTrans = 400 ' 0.4 m/sec
  velocityMaxRot   = 120 ' 120 deg/sec (A4)

  AccelerationMaxTrans = 5000
  AccelerationMaxRot   = 1000 

  JerkMaxTrans = 15*AccelerationMaxTrans 
  JerkMaxRot   = 15*AccelerationMaxRot   

  velocityTrans = 0.1*velocityMaxTrans
  velocityRot   = 0.1*velocityMaxRot

  AccelerationTrans = 0.1*AccelerationMaxTrans 
  AccelerationRot   = 0.1*AccelerationMaxRot

  DecelerationTrans = 0.1*AccelerationMaxTrans 
  DecelerationRot   = 0.1*AccelerationMaxRot

  JerkTrans = 0.1*JerkMaxTrans
  JerkRot   = 0.1*JerkMaxRot

  PositionErrorSettle = 0.025 ' 25um
    PositionErrorMax = 1
    j1.PMin = -180
    j2.PMin = -180
    j3.PMin = -115
    j4.PMin = -190
    j5.PMin = -180

    a1.PMin = -180
    a2.PMin = -180
    a3.PMin = -115
    a4.PMin = -190
    a5.PMin = -180

    j1.PMax = 180
    j2.PMax = 180
    j3.PMax = 115
    j4.PMax = 190
    j5.PMax = 180

    a1.PMax = 180
    a2.PMax = 180
    a3.PMax = 115
    a4.PMax = 190
    a5.PMax = 180
    PositionErrorDelay = GlobalPositionErrorDelay
    detach
  end with
end sub


Function Max(byval a as double,byval b as double ) as double
 If a > b Then
   Max = a
 Else
   Max = b
 End If 
End Function

' Checking VCMD, the function is called every sample so we can do numeric differentiation
Function CheckVel as long

  dim VelCmd as joint of xyzrp
  dim diff   as double

  with ukin
    VelCmd  = 1000*(pcmd - PrevPos)/period ' Computed Velocity  
    PrevVel = vcmd
    PrevPos = pcmd

    diff = Dist(VelCmd,PrevVel)  ' Difference with generated velocity
    if (itype = 1) then
      diff = diff/100   ' In cases of joint interpolation, the vcmd is directly computed by profiler
      diff = 0
    end if              ' So it differs more then in Cartesian interpolation.

    CheckVel = (diff > 1.5)
  end with
end function


function Dist(byval a as joint of xyzrp, byval b as joint of xyzrp) as double
    dim i      as long
    dim diff   as double
    dim d      as double
    diff = 0
    for i = 1 to 5
      d = a{i} - b{i}
      diff = diff + d*d
    next
    Dist = sqrt(diff)
end function

